/* This file is part of SWIS
   Copyright (C) 2007 Sergey Poznyakoff
 
   SWIS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SWIS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SWIS.  If not, see <http://www.gnu.org/licenses/>. */

#include "swis.h"

enum {
  PROGNAME_OPTION,
  EXCLUDE_OPTION,
  EXCLUDE_FROM_OPTION
};

struct option options[] = {
  { "progname", required_argument, NULL, PROGNAME_OPTION },
  { "debug", no_argument, NULL, 'd' },
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { "output", required_argument, NULL, 'o' },
  { "tag", no_argument, NULL, 't' },
  { "files-from", required_argument, NULL, 'T' },
  { "null", no_argument, NULL, '0' },
  { "min-length", required_argument, NULL, 'm' },
  { "badwords", required_argument, NULL, 'b' },
  { "full-text", no_argument, NULL, 'f' },
  { "exclude", required_argument, NULL, EXCLUDE_OPTION },
  { "exclude-from", required_argument, NULL, EXCLUDE_FROM_OPTION },
  { NULL }
};

void
usage ()
{
  printf ("Usage: word-split [OPTIONS] [FILES...]\n");
  printf ("Split UTF-8 input into words\n");
  printf ("\nOptions are:\n");
  printf ("  -b, --badword=FILE             read list of `bad words' from FILE\n");
  /*  printf ("  -d, --debug                    output debugging info\n"); */
  printf ("  -o, --output=FILE              direct output to FILE instead of stdout\n");
  printf ("      --exclude=REGEX            exclude words matching REGEX\n");
  printf ("      --exclude-from=FILE        read list of regexps to exclude from FILE\n");
  printf ("  -m, --min-length=NUMBER        set minimal word length\n");
  printf ("  -t, --tag                      preserve file name tags\n");
  printf ("  -T, --from-file=FILE           read input file names from FILE\n");
  printf ("  -0, --null                     -T reads null-terminated names\n");
  printf ("  -f, --full-text                print full text before the word list\n");
  printf ("\n");
  printf ("  -h, --help                     print this help list\n");
  printf ("  -v, --version                  print program version and exit\n");
  printf ("\n");
  printf ("Report bugs to <%s>\n", PACKAGE_BUGREPORT);
}

char **input_file;
FILE *input;
FILE *output;
int tag_option;
size_t min_length;
int full_text_option;

int
open_input ()
{
  if (input_file && *input_file)
    {
      char *name = *input_file++;
      if (name[0] == '-' && name[1] == 0)
        input = stdin;
      else
        {
          input = fopen (name, "r");
          if (!input)
            error (1, errno, "cannot open input file %s", name);
        }
      return 0;
    }
  return 1;
}

struct obstack stk;
size_t wordcount;

int
string_comp (const void *a, const void *b)
{
  return strcmp (a, b);
}

void
flush_stack ()
{
  char *start, *p;
  char **wtab;
  size_t i;
  size_t offset = 0;
  
  if (!wordcount)
    return;
  start = obstack_finish (&stk);
  wtab = xcalloc (wordcount, sizeof wtab[0]);
  for (i = 0, p = start; i < wordcount; i++, p += strlen (p) + 1)
    wtab[i] = p;

  if (full_text_option)
    {
      fputs (">*", output);
      for (i = 0; i < wordcount; i++)
	{
	  fputs (wtab[i], output);
	  fputc (' ', output);
	}
      fputc ('\n', output);
    }

  offset = 0;
  for (i = 0; i < wordcount; i++)
    {
      fprintf (output, "%lu ", (unsigned long) offset);
      fputs (wtab[i], output);
      fputc ('\n', output);
      offset += utf8_strlen (wtab[i]) + (i > 0);
    }

  free (wtab);
  obstack_free (&stk, start);
  wordcount = 0;
}

int after_delim_output = 1;

void
flush_word (const unsigned *wordbuf)
{
  size_t i;
  size_t wordlen;
  char *wbuf;
  size_t wblen;
  size_t wbc;
  
  if (!wordbuf)
    return;
  
  wordlen = mbutf8_strlen (wordbuf);
  
  if (wordlen <= min_length)
    return;

  wblen = wordlen * 4;
  wbuf = xmalloc (wblen + 1);
  
  wbc = 0;
  for (i = 0; i < wordlen; i++)
    {
      char r[4];
      int rc = utf8_wctomb (r, wordbuf[i]);
      if (rc <= 0)
	error (1, errno, "unexpected error converting UTF-8 string");
      memcpy (wbuf + wbc, r, rc);
      wbc += rc;
    }
  wbuf[wbc++] = 0;

  if (!excluded_word_p (wbuf))
    {
      obstack_grow (&stk, wbuf, wbc);
      wordcount++;
    }
  free (wbuf);
  
  after_delim_output = 1;
}


int
skip_tag (void *data, FILE *fp, int c)
{
  int *pv = data;
  if (tag_option)
    {
      if (*pv && c == '>')
	{
	  flush_stack ();
	  do
	    fputc (c, output);
	  while ((c = fgetc (fp)) != EOF && c != '\n');
	  fputc ('\n', output);
	  after_delim_output = 1;
	  *pv = 0;
	  return 1;
	}
      *pv = c == '\n';
    }
  return 0;
}

int
word_split ()
{
  const unsigned *w;
  int after_newline;
  
  while ((w = fget_word (input, skip_tag, &after_newline)))
    {
      after_delim_output = 0;
      if (!badword_p (w))
	flush_word (w);
    }
  flush_stack ();
  return !open_input ();
}

int
main (int argc, char **argv)
{
  int c;
  char *p;
    
  program_name = argv[0];

  while ((c = getopt_long (argc, argv, "0b:dfm:ho:T:tuv", options, NULL)) != EOF)
    {
      switch (c)
	{
	case '0':
	  filename_terminator = 0;
	  break;

	case 'b':
	  read_badwords (optarg);
	  break;

	case 'd':
	  error (0, 0, "warning: the --debug option is not yet supported");
	  /* FIXME */
	  break;

	case EXCLUDE_OPTION:
	  register_exclude_regexp (optarg);
	  break;

	case EXCLUDE_FROM_OPTION:
	  read_exclude_regexps (optarg);
	  break;
	  
	case 'f':
	  full_text_option = 1;
	  break;
	  
	case 'm':
	  min_length = strtoul (optarg, &p, 10);
	  if (*p)
	    error (1, 0, "invalid number (near %s)", p);
	  break;
	      
	case 'h':
	  usage ();
	  exit (0);

	case 'o':
	  output = fopen (optarg, "w");
	  if (!output)
	    error (1, errno, "cannot open output file %s", optarg);
	  break;

	case PROGNAME_OPTION:
	  program_name = optarg;
	  break;
	  
	case 'T':
	  read_names_from_file (optarg);
	  break;
	  
	case 't':
	  tag_option = 1;
	  break;

	case 'v':
	  swis_version (stdout, "word-split");
	  exit (0);
	  
	default:
	  exit (1);
	}
    }

  argc -= optind;
  argv += optind;

  update_argcv (&argc, &argv);

  if (argc)
    {
      input_file = argv;
      open_input ();
    }
  
  if (!input)
    input = stdin;
  if (!output)
    output = stdout;

  obstack_init (&stk);
  
  while (word_split ())
    ;
  
  exit (0);
}

 

  
